import logging
import multiprocessing
from concurrent import futures
from concurrent.futures import Executor
from tempfile import TemporaryDirectory
from typing import Dict, Any, Type, Optional

from kubernetes.client import ApiException
from kubernetes.client import Configuration

import fixlib.logger
import fixlib.proc
from fix_plugin_k8s.base import K8sApiClient, K8sClient
from fix_plugin_k8s.base import K8sConfig
from fix_plugin_k8s.collector import KubernetesCollector
from fix_plugin_k8s.deferred_edges import create_deferred_edges
from fixlib.args import ArgumentParser, Namespace
from fixlib.baseplugin import BaseCollectorPlugin
from fixlib.config import Config, RunningConfig
from fixlib.core.actions import CoreFeedback
from fixlib.graph import Graph

log = logging.getLogger("fix.plugins.k8s")


class KubernetesCollectorPlugin(BaseCollectorPlugin):
    cloud = "k8s"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # once defined, it will be set by the worker
        self.core_feedback: Optional[CoreFeedback] = None

    def collect(self, **kwargs: Any) -> None:
        log.debug("plugin: Kubernetes collecting resources")
        assert self.core_feedback, "core_feedback is not set"

        k8s: K8sConfig = Config.k8s
        with TemporaryDirectory() as tmpdir:
            cluster_access = k8s.cluster_access_configs(tmpdir, self.core_feedback)

            if len(cluster_access) == 0:
                log.warning("Kubernetes plugin enabled, but no clusters configured. Ignore.")
                return

            max_workers = len(cluster_access) if len(cluster_access) < k8s.pool_size else k8s.pool_size
            pool_args: Dict[str, Any] = {"max_workers": max_workers}
            if k8s.fork_process:
                pool_args["mp_context"] = multiprocessing.get_context("spawn")
                pool_args["initializer"] = fixlib.proc.collector_initializer
                pool_executor: Type[Executor] = futures.ProcessPoolExecutor
            else:
                pool_executor = futures.ThreadPoolExecutor

            with pool_executor(**pool_args) as executor:
                wait_for = [
                    executor.submit(
                        self.collect_cluster,
                        cluster_id,
                        cluster_config,
                        ArgumentParser.args,
                        Config.running_config,
                        self.core_feedback.with_context("k8s", cluster_id),
                        **kwargs,
                    )
                    for cluster_id, cluster_config in cluster_access.items()
                ]
                for future in futures.as_completed(wait_for):
                    cluster_graph = future.result()
                    if not isinstance(cluster_graph, Graph):
                        log.error(f"Skipping invalid cluster_graph {type(cluster_graph)}")
                        continue
                    self.send_account_graph(cluster_graph)

    @staticmethod
    def collect_cluster(
        cluster_id: str,
        cluster_config: Configuration,
        args: Namespace,
        running_config: RunningConfig,
        core_feedback: CoreFeedback,
        **kwargs: Any,
    ) -> Graph:
        """
        Collects an individual Kubernetes Cluster.
        """
        fixlib.proc.set_thread_name(f"k8s_{cluster_id}")

        if args is not None:
            ArgumentParser.args = args
        if running_config is not None:
            Config.running_config.apply(running_config)

        log.debug(f"Starting new collect process for {cluster_id}")

        try:
            k8s_client: K8sClient = kwargs.get("client_factory", K8sApiClient.from_config)(
                cluster_id, cluster_config
            ).with_feedback(core_feedback)
            kc = KubernetesCollector(Config.k8s, k8s_client)
            kc.collect()
            create_deferred_edges(kc.graph)
        except ApiException as e:
            if e.reason == "Unauthorized":
                core_feedback.error(f"Unable to authenticate with {cluster_id}", log)
            else:
                core_feedback.error(f"An unhandled error occurred while collecting {cluster_id}: {e}", log)
            raise
        except Exception as e:
            core_feedback.error(f"An unhandled error occurred while collecting {cluster_id}: {e}", log)
            raise
        else:
            return kc.graph

    @staticmethod
    def add_config(config: Config) -> None:
        config.add_config(K8sConfig)



from fixlib.baseresources import BaseResource
from fixlib.graph import BySearchCriteria, ByNodeId, Graph
from fix_plugin_k8s.deferred_edges.utils import rgetattr


def link_k8s_node_to_vmss_instance(graph: Graph, resource: BaseResource) -> None:
    if resource.kind == "kubernetes_node":
        if (pid := rgetattr(resource, "node_spec.provider_id", None)) and (pid.startswith("azure://")):
            vmss_vmss_instance_id = pid.replace("azure://", "", 1)
            graph.add_deferred_edge(
                BySearchCriteria(
                    f"is(azure_virtual_machine_scale_set_instance) and reported.id={vmss_vmss_instance_id}"
                ),
                ByNodeId(resource.chksum),
            )


def link_k8s_cluster_to_aks_cluster(graph: Graph, resource: BaseResource) -> None:
    if resource.kind == "kubernetes_cluster":
        graph.add_deferred_edge(
            BySearchCriteria(f"is(azure_managed_cluster) and reported.name={resource.id}"),
            ByNodeId(resource.chksum),
        )


def link_service_to_azure_lb(graph: Graph, resource: BaseResource) -> None:
    if resource.kind == "kubernetes_service":
        if (rgetattr(resource, "service_spec.type", None) == "LoadBalancer") and (
            ingresses := rgetattr(resource, "service_status.load_balancer.ingress", None)
        ):
            for ingress in ingresses:
                if lb_ip := rgetattr(ingress, "ip", None):
                    graph.add_deferred_edge(
                        BySearchCriteria(f"is(azure_load_balancer) and reported.aks_public_ip_address={lb_ip}"),
                        ByNodeId(resource.chksum),
                    )


def link_pv_to_azure_disk(graph: Graph, resource: BaseResource) -> None:
    if resource.kind == "kubernetes_persistent_volume":
        if (
            (csi := rgetattr(resource, "persistent_volume_spec.csi", None))
            and (csi.get("driver") == "disk.csi.azure.com")
            and (vol_id := csi.get("volumeHandle"))
        ):
            graph.add_deferred_edge(
                BySearchCriteria(f"is(azure_disk) and reported.id={vol_id}"), ByNodeId(resource.chksum)
            )


def link_all(graph: Graph, resource: BaseResource) -> None:
    link_k8s_node_to_vmss_instance(graph, resource)
    link_k8s_cluster_to_aks_cluster(graph, resource)
    link_service_to_azure_lb(graph, resource)
    link_pv_to_azure_disk(graph, resource)



from fixlib.baseresources import BaseResource
from fixlib.graph import BySearchCriteria, ByNodeId, Graph
from fix_plugin_k8s.deferred_edges.utils import rgetattr


def link_node_to_do_droplet(graph: Graph, resource: BaseResource) -> None:
    if resource.kind == "kubernetes_node":
        if (pid := rgetattr(resource, "node_spec.provider_id", None)) and (pid.startswith("digitalocean://")):
            _, droplet_id = pid.split("digitalocean://")
            graph.add_deferred_edge(
                BySearchCriteria(f"is(digitalocean_droplet) and reported.id={droplet_id}"),
                ByNodeId(resource.chksum),
            )


def link_service_to_do_lb(graph: Graph, resource: BaseResource) -> None:
    if resource.kind == "kubernetes_service":
        if lb_id := resource.tags.get("kubernetes.digitalocean.com/load-balancer-id"):
            graph.add_deferred_edge(
                BySearchCriteria(f"is(digitalocean_load_balancer) and reported.id={lb_id}"),
                ByNodeId(resource.chksum),
            )


def link_pv_to_do_volume(graph: Graph, resource: BaseResource) -> None:
    if resource.kind == "kubernetes_persistent_volume":
        if (
            (csi := rgetattr(resource, "persistent_volume_spec.csi", None))
            and (csi.get("driver") == "dobs.csi.digitalocean.com")
            and (vol_id := csi.get("volumeHandle"))
        ):
            graph.add_deferred_edge(
                BySearchCriteria(f"is(digitalocean_volume) and reported.id={vol_id}"), ByNodeId(resource.chksum)
            )


def link_all(graph: Graph, resource: BaseResource) -> None:
    link_node_to_do_droplet(graph, resource)
    link_service_to_do_lb(graph, resource)
    link_pv_to_do_volume(graph, resource)
